# Copyright (c) 2014, Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
r"""
    pkgimp - Python Java-style Package imports
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    .. license:: MIT License
    .. author::  Niklas Rosenstein <rosensteinniklas@gmail.com>
    .. version:: 0.0.3
"""

import os
import sys
import imp

__version__ = (0, 0, 3)
__author__ = 'Niklas Rosenstein <rosensteinniklas@gmail.com>'

class PackageModuleFinder(object):
    r""" This class implements the Importer protocol introduced with
    `PEP 302`_. It extends the Python import functionality by Java-style
    import capabilities.

    It uses :attr:`sys.path` to find the module that is to be
    imported but will accept directory names as package identifiers.

    For instance, when the following directory structures can be found
    on the path (of course existing in two separate directories, otherwise
    the directories would need to be merged)

    .. code-block::

        org/
            lang/
                __init__.py
                exceptions.py

        org/
            runtime.py

    The following imports are valid after the :class:`PackageModuleFinder`
    was installed on :attr:`sys.meta_path`.

    .. code-block:: python

        import org.lang
        import org.lang.exceptions
        import org.runtime

    .. _PEP 302: http://www.python.org/dev/peps/pep-0302
    """

    def __init__(self, prefix=None, stripprefix=False):
        r""" Initialize the PackageModuleFinder.

        :param prefix: This optional parameter can be a string that
                is a package identifier (separated by dots) that must
                be given as a prefix for importing Java-style packages.
        :param stripprefix: If this is True, the prefix will be removed
                before resolving the package name.
        """

        super(PackageModuleFinder, self).__init__()
        self.prefix = prefix.split('.') if prefix else []
        self.stripprefix = stripprefix

    def resolve_import(self, dirname, parts):
        for part in parts[:-1]:
            dirname = os.path.join(dirname, part)
            if not os.path.isdir(dirname):
                return None

        directory = os.path.join(dirname, parts[-1])
        modulefile = os.path.join(dirname, '%s.py' % parts[-1])
        initfile = os.path.join(directory, '__init__.py')

        if os.path.isfile(modulefile):
            return modulefile
        elif os.path.isfile(initfile):
            return initfile
        elif os.path.isdir(directory):
            return directory
        else:
            return None

    def find_module(self, fullname, path=None):
        # Default to the system path if no path was specified.
        if path is None:
            path = sys.path

        # Split the fullname into its sub-module names that
        # will be used to resolve the module name.
        parts = fullname.split('.')

        # If there is a prefix specified in the PackageModuleFinder,
        # does the current import match the prefix?
        if self.prefix:
            for a, b in zip(self.prefix, parts):
                if a != b:
                    return None
            if self.stripprefix:
                parts = parts[len(self.prefix):]

                # If there are no parts left of the module which is
                # to be imported, the prefix matches the fullname exactly.
                if not parts:
                    return FakePackageImporter()

        # Is there a directory on the path that qualifies with
        # the specified import names?
        filepath = None
        for dirname in path:
            filepath = self.resolve_import(dirname, parts)
            if filepath:
                break

        if filepath:
            if os.path.isdir(filepath):
                return FakePackageImporter()
            else:
                return RealModuleImporter(filepath)

        return None

class FakePackageImporter(object):

    def load_module(self, fullname):
        mod = sys.modules.get(fullname)
        if mod:
            return mod

        mod = imp.new_module(fullname)
        sys.modules[fullname] = mod
        mod.__file__ = '<%s>' % fullname
        mod.__loader__ = self
        mod.__path__ = sys.path
        mod.__package__ = fullname
        return mod

class RealModuleImporter(object):

    def __init__(self, filename):
        r""" Create a RealModuleImporter for a specified filename. """
        super(RealModuleImporter, self).__init__()
        self.filename = filename

    def get_code(self):
        with open(self.filename) as fp:
            source = fp.read()

        pre = 'from __future__ import absolute_import;del absolute_import;'
        return compile(pre + source, self.filename, 'exec')

    def load_module(self, fullname):
        mod = sys.modules.get(fullname)
        if mod:
            return mod

        mod = imp.new_module(fullname)
        sys.modules[fullname] = mod
        mod.__file__ = self.filename
        mod.__loader__ = self

        basename = os.path.basename(self.filename)
        if basename == '__init__.py':
            mod.__path__ = sys.path
            mod.__package__ = fullname
        else:
            mod.__package__ = fullname.rpartition('.')[0]

        code = self.get_code()
        exec code in mod.__dict__

        return mod

def install(prefix, stripprefix=False, __cache={}):
    r""" Installs a :class:`PackageModuleFinder` on :attr:`sys.meta_path`
    if it was not already installed. """

    if prefix not in __cache:
        finder = PackageModuleFinder(prefix, stripprefix)
        sys.meta_path.append(finder)
        __cache[prefix] = finder

